#enigma.py
#Enigma Machine Simulator. Version 1.4
#Copyright (c) 2020, Julie VK3FOWL and Joe VK3YSP
#For the School Amateur Radio Club Network https://www.sarcnet.org/
#This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
#This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
    
#This program illustrates the design of an enigma machine simulator written in Python 3
#It can simulate the enigma I, M3 and M4 machines

#References
#https://en.wikipedia.org/wiki/Enigma_machine
#https://en.wikipedia.org/wiki/Cryptanalysis_of_the_Enigma

#Here is brief recap of how enigma works. See the above references for more details:
#When one of 26 letters (A-Z) is pressed on the keyboard, it is scrambled progressively by:
#   The plugboard wiring;
#   3-4 rotors, each with specific wiring, in the forward, or right-to-left, direction;
#   the static reflector wiring;
#   the same 3-4 rotors, but in the reverse, or left-to-right, direction;
#   and finally, the plugboard wiring again.
#Each stage is a simple substitution cipher: The rotors have 26 substitutions, the plugboard has 13 pairs of substitutions
#The message is further scrambled using the following methods:
#   Inserting cables into the plugboard (usually 10 out of a possible 13). Each cable creates a symmetrical substitution of keys and lamps.
#   Setting the rotor starting positions
#   Offsetting the rotor positions with a ring
#   When a button is pressed, but before any output is displayed:
#       If any rotor is at one or two predetermined notch positions on its ring, the rotor to its left is advanced one step
#       The right hand rotor is unconditionally advanced one step
#The output letter (A-Z) is then illuminated on one of 26 lamps
#The output changes each time a key is pressed, even if it is the same key. However, the output cannot be the same as the keyed letter
#There is no space. Uncommon letters, like X, may be used instead.
#The enigma machine can be used symmetrically, for either encryption or description, with exactly the same settings

#Version History
#1.0    Initial Release
#1.1    Fixed 'double stepping'
#1.2    Object Oriented version with test messages for enigma class
#1.3    Support for common user formats for rotor order, message key, reflector and plug pairs 
#1.4    Added more test messages
#1.5    Added source URL to output

#TODO:
#Error checking
#Discover the problem with the Scharnhorst message
#Discover the problem with PAGE_40_PYCMW (Norrköping Enigma message)
#More testing with authentic messages
#Testing with 5-rotor messages

class Enigma(object):
    def __init__(self):
        #Initialise instance variables

        #We first define the static settings of the machine as lists of strings, which we have preserved in the program for verification with historical sources
        #We found this information at: https://en.wikipedia.org/wiki/Enigma_rotor_details

        #The forward (right-to-left) wiring for each rotor type (I, II, III, IV, V, VI, VII, VIII, beta, gamma)
        self.rotorForwardWiringStringList = ['EKMFLGDQVZNTOWYHXUSPAIBRCJ', 'AJDKSIRUXBLHWTMCQGZNPYFVOE', 'BDFHJLCPRTXVZNYEIWGAKMUSQO',
                                            'ESOVPZJAYQUIRHXLNFTGKDCMWB', 'VZBRGITYUPSDNHLXAWMJQOFECK', 'JPGVOUMFYQBENHZRDKASXLICTW',
                                            'NZJHGRCXMYSWBOUFAIVLPEKQDT', 'FKQHTLXOCBJSPDZRAMEWNIUYGV','LEYJVCNIXWPBQMDRTAKZGFUHOS',
                                            'FSOKANUERHMBTIYCWLQPZXVGJD']
        
        self.rotorDictionary = {'I':0, 'II':1, 'III':2, 'IV':3, 'V':4, 'VI':5, 'VII':6, 'VIII':7, 'BETA':8, 'GAMMA':9}
        
        #The reflector wiring for each reflector type (A, B, C, B Thin, C Thin)
        self.reflectorWiringStringList = ['EJMZALYXVBWFCRQUONTSPIKHGD', 'YRUHQSLDPXNGOKMIEBFZCWVJAT', 'FVPJIAOYEDRZXWGCTKUQSBNMHL',
                                         'ENKQAUYWJICOPBLMDXZVFTHRGS', 'RDOBJNTKVEHMLFCWZAXGYIPSUQ']

        self.reflectorDictionary = {'A':0, 'B':1, 'C':2, 'B THIN':3, 'C THIN':4}

        #The notch position(s) for each rotor type (I, II, III, IV, V, VI, VII, VIII). Note: These refer to ring positions, not to the rotor positions. 
        self.notchPositionStringList = ['Q', 'E', 'V', 'J', 'Z', 'ZM', 'ZM', 'ZM', 'ZM', 'ZM']

        #Default Simulator Settings

        #Reflector type used (Note: Reflector types A, B, C, Thin B, Thin C are coded 0..5)
        #Rotor types used in each rotor position (Note: Rotor types I..VIII are coded 1-8, 0 for no rotor used)
        #Ring settings used in each rotor position (Note: Valid ring positions are: 1..26, 0 for no rotor used)
        #Start positions used in each rotor position (Note: Valid rotor positions: A..Z or 1..26, 0 for no rotor used) 

        self.name =         'Operation Barbarossa, 1941 Pt 1'
        self.url =          'http://wiki.franklinheath.co.uk/index.php/Enigma/Sample_Messages'          
        self.plaintext =    'AUFKL XABTE ILUNG XVONX KURTI NOWAX KURTI NOWAX NORDW ESTLX SEBEZ XSEBE ZXUAF FLIEG ERSTR ASZER IQTUN GXDUB ROWKI XDUBR OWKIX OPOTS CHKAX OPOTS CHKAX UMXEI NSAQT DREIN ULLXU HRANG ETRET ENXAN GRIFF XINFX RGTX'
        self.ciphertext =   'EDPUD NRGYS ZRCXN UYTPO MRMBO FKTBZ REZKM LXLVE FGUEY SIOZV EQMIK UBPMM YLKLT TDEIS MDICA GYKUA CTCDO MOHWX MUUIA UBSTS LRNBZ SZWNR FXWFY SSXJZ VIJHI DISHP RKLKA YUPAD TXQSP INQMA TLPIF SVKDA SCTAC DPBOP VHJK'
        self.rotorOrder =               ['II', 'IV', 'V']   #Rotor order: II IV V
        self.ringList =                 [2,21,12]           #Ring positions: 02 21 12
        self.startString =              'BLA'               #Start positions as string: BLA;
        self.reflectorName =            'B'
        self.plugboardWiringString =    'AV BS CG DL FU HZ IN KM OW RX'
        #Initialization
        self.initialize()

    #Methods

    #The enigma machine wiring and settings is published as strings of capital letters.
    #We want to preserve this information as-is for historical and operational reasons, but we also want our code to be efficient
    #So we have to do a lot of conversions from strings to lists of zero-based integers and vice verse.
    #Note that each character of the strings is actually encoded in ASCII, such that capital A = 65, B = 66 and so on.
    #To make efficient zero-based integers from these characters, we have to subtract 65 from the ordinal value of each ASCII character such that A = 0, B = 1 etc
    #Therefore, string lists of the form ['ABC..XYZ','ZYX..CBA'] become integer lists like [[0,1,2..23,24,25],[25,24,23..2,1,0]] etc 
    #We will define the following functions to do the conversion

            
    def printHeading(self):
        #Print a heading for the program
        print('Enigma Machine Simulator. Version 1.5')
        print('Copyright (c) 2020, Julie VK3FOWL and Joe VK3YSP')
        print('For the School Amateur Radio Club Network https://www.sarcnet.org/')
        print('This is free software released into the public domain under the GNU General Public License')
        print()

    def initialize(self):
        #Initialization - To be run after all settings are entered, but before processing.
        #Infer the number of rotors from the rotorOrder list
        self.rotors = len(self.rotorOrder)
        #Create a zero-based integer list of the rotor types being used from the specified rotor order list
        self.rotorTypes = self.rotorOrderToRotorTypes(self.rotorOrder)
        #Create a zero-based integer list of the ring positions from the specified ring list
        self.ringPositions = self.ringListToRingPositions(self.ringList) 
        #Create a reflector types list from reflector names list
        self.reflectorType = self.reflectorDictionary[self.reflectorName]
        #Create start positions from start string
        self.startPositions = self.startStringToStartPositions(self.startString)
        self.rotorPositions = self.startPositionsToRotorPositions(self.startPositions, self.ringPositions)
        #Create the rotor forward and reverse wiring lists
        self.rotorForwardWiring = self.asciiStringsToIntegerLists(self.rotorForwardWiringStringList)
        self.rotorReverseWiring = self.reverseRotorForwardWiringIntegerLists(self.rotorForwardWiring)
        #Create the reflector wiring list
        self.reflectorWiringIntegerLists = self.asciiStringsToIntegerLists(self.reflectorWiringStringList)
        self.reflectorWiring = self.reflectorWiringIntegerLists[self.reflectorDictionary[self.reflectorName]]
        #Create the notch positions list
        self.notchPositions = self.asciiStringsToIntegerLists(self.notchPositionStringList)
        #Create the plugboard wiring list
        self.plugboardWiring = self.plugboardWiringStringToPlugboardWiringList(self.plugboardWiringString.replace(' ',''))
        #Remove spaces from the plaintext and ciphertext
        self.plaintext = self.plaintext.replace(' ','')
        self.ciphertext = self.ciphertext.replace(' ','')

    def rotorOrderToRotorTypes(self, stringList):
        #Converts a rotor order string list into list of zero-based rotor types, e.g. ['II', 'IV', 'V'] into [2,4,5,0]
        rotorTypes = [0,0,0,0]
        for i, string in enumerate(stringList):
            rotorTypes[i] = self.rotorDictionary[string]
        return rotorTypes

    def ringListToRingPositions(self, ringList):
        #Converts a one-based ring list to a zero-based ring positions list, e.g. [2, 21, 12] into [1, 20, 11, 0]
        #Converts a ring position string to a zero-based ring positions list, e.g. 'BUL' into [1, 20, 11, 0]
        ringPositions = [0,0,0,0]
        if isinstance(ringList, list):
            for i, integer in enumerate(ringList):
                ringPositions[i] = integer - 1
        if isinstance(ringList, str):
            for i, character in enumerate(ringList):
                ringPositions[i] = self.ordinal(character)
        return ringPositions

    def integerListToAsciiString(self, integerList):
        #Converts a zero-based integer list to a string of ascii characters
        asciiString = ''
        for integer in integerList:
            asciiString += self.ascii(integer)
        return asciiString
        
    def startStringToStartPositions(self, string):
        #Converts a startString to a zero-based start positions list, e.g. 'BLA' into [1, 12, 0, 0]
        startPositions = [0,0,0,0]
        for i, character in enumerate(string):
            startPositions[i] = self.ordinal(character)
        return startPositions

    def ascii(self, integer):
        #Converts a zero-based integer (0-25) into an ASCII character (A..Z)
        return chr(integer + 65)

    def ordinal(self, character):
        #Converts an ASCII character (A..Z) into a zero-based integer (0-25)
        return ord(character) - 65

    def asciiStringToIntegerList(self, asciiString):
        #Converts a single ASCII string into a list of zero-based integers, e.g. 'ABCXYZ' into [0, 1, 2, 23, 24, 25]
        integerList = []                                                #Clear the list
        for character in asciiString:                                   #Iterate through each character in the string
            integerList.append(self.ordinal(character))                      #Get the ordinal value of each ASCII character and append it to the array
        return integerList                                              #Return the list

    def asciiStringsToIntegerLists(self, asciiStrings):
        #Converts a list of ASCII strings into lists of zero-based integers, e.g. ['AB', 'CX', 'YZ'] into [[0, 1], [2, 23], [24, 25]] 
        integerLists = []                                               #Clear the list
        for asciiString in asciiStrings:                                #Iterate through each string in the list
            integerList = self.asciiStringToIntegerList(asciiString)         #Get the integer list from the ASCII strings
            integerLists.append(integerList)                            #Append the integer list
        return integerLists                                             #Return the list of integer lists

    def integerListOneBaseToIntegerListZeroBase(self, integerListOne):
        #Converts a one-based integer list into a zero-based integer list, e.g. [1, 2, 25, 26] into [0, 1, 24, 25]
        IntegerListZero = []
        for integer in integerListOne:
            IntegerListZero.append(integer - 1)
        return IntegerListZero

    def plugboardWiringStringToPlugboardWiringList(self, plugboardWiringString):
        #Converts plugboard wiring strings into integers e.g.
        #'AVBSCGDLFUHZINKMOWRX' into [21,18,6,11,4,20,2,25,13,9,12,3,10,8,22,15,16,23,1,19,5,0,14,17,24,7]
        PlugboardWiringIntegerList = [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25]
        while len(plugboardWiringString) > 0:
            string = plugboardWiringString[0:2]
            PlugboardWiringIntegerList[self.ordinal(string[0])] = self.ordinal(string[1])
            PlugboardWiringIntegerList[self.ordinal(string[1])] = self.ordinal(string[0])
            plugboardWiringString = plugboardWiringString[2:]
        return PlugboardWiringIntegerList

    def plugboardWiringStringListToPlugboardWiringIntegerList(self, plugboardWiringStringList):
        #Converts plugboard wiring string lists into integers e.g.
        #['AV','BS','CG','DL','FU','HZ','IN','KM','OW','RX'] into [21,18,6,11,4,20,2,25,13,9,12,3,10,8,22,15,16,23,1,19,5,0,14,17,24,7]
        PlugboardWiringIntegerList = [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25]
        for string in plugboardWiringStringList:
            PlugboardWiringIntegerList[self.ordinal(string[0])] = self.ordinal(string[1])
            PlugboardWiringIntegerList[self.ordinal(string[1])] = self.ordinal(string[0])
        return PlugboardWiringIntegerList

    def mod26(self, value):
        #Get the cyclic position on a 26-step rotor after adding or subtracting any offset e.g. mod26(10)=10, mod26(-1)=25, mod26(26)=0, mod26(27)=1 
        return value % 26

    def reverseRotorForwardWiringIntegerLists(self, integerLists):
        #Reverses the lookup direction of the integer lists - used for reversing the direction of rotor wiring lists
        #e.g. converts [[2, 3, 0, 1], [3, 0, 1, 2]] into [[2, 3, 0, 1], [1, 2, 3, 0]]
        reverseIntegerLists = []                                        #Clear the list of lists
        for integerList in integerLists:                                #Iterate through each list in the lists
            reverseIntegerList = []                                     #Clear the list
            for i in range(len(integerList)):                           #Iterate through each integer in the integer list 
                reverseIntegerList.append(integerList.index(i))         #Replace each integer with its own lookup value
            reverseIntegerLists.append(reverseIntegerList)              #Append the new reversed integer list to the list of lists
        return reverseIntegerLists                                      #Return the list of reversed integer lists

    def startPositionsToRotorPositions(self, startPositions, ringPositions):
        #Converts the displayed rotor start positions given the internal rotor ring positions into the actual rotor wiring positions
        # e.g. [1, 11, 0, -1] - [1, 20, 11, -1] = [0, 17, 15, 0]
        rotorPositions = []
        for i, start in enumerate(startPositions):
            rotorPositions.append(self.mod26(start - ringPositions[i]))
        return rotorPositions

    def process(self, message):
        #Symmetrically encodes or decodes a message using the enigma machine settings
        result = ''
        for character in message:
            result += self.ascii(self.enigma(self.rotors,self.ordinal(character)))   #Select only one of these. TODO: Make this user-friendly
        return result

    def step(self, rotors):
        #Rotor stepping function for 3 or 4 rotors
        step3 = (self.mod26(self.rotorPositions[3] + self.ringPositions[3]) in self.notchPositions[self.rotorTypes[3]]) #Need to advance rotor 3
        step2 = (self.mod26(self.rotorPositions[2] + self.ringPositions[2]) in self.notchPositions[self.rotorTypes[2]]) #Need to advance rotor 2
        step1 = (self.mod26(self.rotorPositions[1] + self.ringPositions[1]) in self.notchPositions[self.rotorTypes[1]]) #Need to advance rotor 1
        if step1: self.rotorPositions[0] = self.mod26(self.rotorPositions[0] + 1) #Advance rotor 1
        if step2: self.rotorPositions[1] = self.mod26(self.rotorPositions[1] + 1) #Advance rotor 2
        #Unconditionally advance the last rotor
        if (rotors == 3) or step3: self.rotorPositions[2] = self.mod26(self.rotorPositions[2] + 1) #Advance rotor 3
        if (rotors == 4): self.rotorPositions[3] = self.mod26(self.rotorPositions[3] + 1) #Advance rotor 4
        
    def rotor(self, wiring, position, value):
        #The transfer function for a single rotor 
        return self.mod26(wiring[self.mod26(value + position)] - position)
        
    def enigma(self, rotors, character):
        #The transfer function for a multi-rotor enigma machine
        self.step(self.rotors)                                              #Advance the rotors before computing the result
        character = self.plugboardWiring[character]                         #Process the character through the input plugboard
        for i in range(rotors - 1, -1, -1):                                 #Process the character through all rotors from right to left
            character = self.rotor(self.rotorForwardWiring[self.rotorTypes[i]], self.rotorPositions[i], character)  #Process the character through one rotor from left to right
        character = self.reflectorWiring[character]                         #Process the character through the reflector
        for i in range(0, rotors, 1):                                       #Process the character through all rotors from left to right
            character = self.rotor(self.rotorReverseWiring[self.rotorTypes[i]], self.rotorPositions[i], character)  #Process the character through one rotor from right to left
        character = self.plugboardWiring[character]                         #Process the character through the output plugboard
        return character

    def print5x5(self, title, string):
        #Print a block of text in 5 columns of 5 characters each
        line = title
        for i, character in enumerate(string):
            line += character
            if i % 5 == 4:
                line += ' '
            if i % 25 == 24:
                print(line)
                line = ''
                for i in range(len(title)): line += ' '
        print(line)

    def printConfiguration(self):
        #Print the complete list of settings
        print('Configuration:')
        print()
        for i, (name, number) in enumerate(self.rotorDictionary.items()):
            print(  f'Rotor Name:                       {name}')
            print(  f'Rotor Forward Wiring:             {self.rotorForwardWiringStringList[number]}')
            print(  f'Rotor Reverse Wiring:             {self.integerListToAsciiString(self.rotorReverseWiring[number])}')
            print(  f'Rotor Notch Positions:            {self.integerListToAsciiString(self.notchPositions[number])}')
            print()
        for i, (name, number) in enumerate(self.reflectorDictionary.items()):
            print(  f'Reflector Name:                   {name}')      
            print(  f'Reflector Wiring:                 {self.reflectorWiringStringList[number]}')
            print()

    def test(self):
        print(f'Message:          {self.name}')
        print(f'Source URL:       {self.url}')
        print(f'Rotor order:      {self.rotorOrder}')
        print(f'Ring positions:   {self.ringList}')
        print(f'Message key:      {self.startString}')
        print(f'Reflector:        {self.reflectorName}')
        print(f'Plug pairs:       {self.plugboardWiringString}')
        self.print5x5('Plaintext:        ', self.plaintext)
        self.print5x5('Ciphertext:       ', self.ciphertext)
        self.result1 = self.process(self.plaintext)
        #We can check the result simply by resetting the rotors to the starting position and processing the ciphertext as a message
        self.rotorPositions = self.startPositionsToRotorPositions(self.startPositions, self.ringPositions)
        self.result2 = self.process(self.ciphertext)
        #self.print5x5('Encoding:         ', self.result1)
        #self.print5x5('Decoding:         ', self.result2)
        if self.result1 == self.ciphertext:
            print('Encoding:         VERIFIED')
        else:    
            print('Encoding:         NOT VERIFIED')
        if self.result2 == self.plaintext:
            print('Decoding:         VERIFIED')
        else:    
            print('Decoding:         NOT VERIFIED')
        print()
 
if __name__ == "__main__":
    #Create an enigma object
    enigma = Enigma()

    enigma.printHeading()
    enigma.printConfiguration() #Comment this out to save on output

    #Enigma machine sample messages from http://wiki.franklinheath.co.uk/index.php/Enigma/Sample_Messages
    
    enigma.name =       'Operation Barbarossa, 1941 Pt 1'   #VERIFIED
    enigma.url =        'http://wiki.franklinheath.co.uk/index.php/Enigma/Sample_Messages'
    enigma.plaintext =  'AUFKL XABTE ILUNG XVONX KURTI NOWAX KURTI NOWAX NORDW ESTLX SEBEZ XSEBE ZXUAF FLIEG ERSTR ASZER IQTUN GXDUB ROWKI XDUBR OWKIX OPOTS CHKAX OPOTS CHKAX UMXEI NSAQT DREIN ULLXU HRANG ETRET ENXAN GRIFF XINFX RGTX'
    enigma.ciphertext = 'EDPUD NRGYS ZRCXN UYTPO MRMBO FKTBZ REZKM LXLVE FGUEY SIOZV EQMIK UBPMM YLKLT TDEIS MDICA GYKUA CTCDO MOHWX MUUIA UBSTS LRNBZ SZWNR FXWFY SSXJZ VIJHI DISHP RKLKA YUPAD TXQSP INQMA TLPIF SVKDA SCTAC DPBOP VHJK'
    enigma.rotorOrder =             ['II', 'IV', 'V']
    enigma.ringList =               [2,21,12]
    enigma.startString =            'BLA'
    enigma.reflectorName =          'B'
    enigma.plugboardWiringString =  'AV BS CG DL FU HZ IN KM OW RX'
    enigma.initialize()
    enigma.test()

    enigma.name =       'Operation Barbarossa, 1941 Pt 2'   #VERIFIED
    enigma.url =        'http://wiki.franklinheath.co.uk/index.php/Enigma/Sample_Messages'
    enigma.plaintext =  'DREIG EHTLA NGSAM ABERS IQERV ORWAE RTSXE INSSI EBENN ULLSE QSXUH RXROE MXEIN SXINF RGTXD REIXA UFFLI EGERS TRASZ EMITA NFANG XEINS SEQSX KMXKM XOSTW XKAME NECXK'
    enigma.ciphertext = 'SFBWD NJUSE GQOBH KRTAR EEZMW KPPRB XOHDR OEQGB BGTQV PGVKB VVGBI MHUSZ YDAJQ IROAX SSSNR EHYGG RPISE ZBOVM QIEMM ZCYSG QDGRE RVBIL EKXYQ IRGIR QNRDN VRXCY YTNJR'
    enigma.rotorOrder =             ['II', 'IV', 'V']
    enigma.ringList =               [2,21,12]
    enigma.startString =            'LSD'
    enigma.reflectorName =          'B'
    enigma.plugboardWiringString =  'AV BS CG DL FU HZ IN KM OW RX'
    enigma.initialize()
    enigma.test()
    
    enigma.name =       'Scharnhorst (Konteradmiral Erich Bey), 1943'   #NOT VERIFIED
    enigma.url =        'http://wiki.franklinheath.co.uk/index.php/Enigma/Sample_Messages'
    enigma.plaintext =  'YKAE NZAP MSCH ZBFO CUVM RMDP YCOF HADZ IZME FXTH FLOL PZLF GGBO TGOX GRET DWTJ IQHL MXVJ WKZU ASTR'
    enigma.ciphertext = 'STEUE REJTA NAFJO RDJAN STAND ORTQU AAACC CVIER NEUNN EUNZW OFAHR TZWON ULSMX XSCHA RNHOR STHCO'
    enigma.rotorOrder =             ['III', 'VI', 'VIII']
    enigma.ringList =               [1,8,13]
    enigma.startString =            'UZV'
    enigma.reflectorName =          'B'
    enigma.plugboardWiringString =  'AN EZ HK IJ LR MQ OT PV SW UX'
    enigma.initialize()
    enigma.test()

    enigma.name =       'Enigma Instruction Manual, 1930'   #VERIFIED
    enigma.url =        'http://wiki.franklinheath.co.uk/index.php/Enigma/Sample_Messages'
    enigma.plaintext =  'FEIND LIQEI NFANT ERIEK OLONN EBEOB AQTET XANFA NGSUE DAUSG ANGBA ERWAL DEXEN DEDRE IKMOS TWAER TSNEU STADT'
    enigma.ciphertext = 'GCDSE AHUGW TQGRK VLFGX UCALX VYMIG MMNMF DXTGN VHVRM MEVOU YFZSL RHDRR XFJWC FHUHM UNZEF RDISI KBGPM YVXUZ'
    enigma.rotorOrder =             ['II', 'I', 'III']
    enigma.ringList =               [24,13,22]
    enigma.startString =            'ABL'
    enigma.reflectorName =          'A'
    enigma.plugboardWiringString =  'AM FI NV PS TU WZ'
    enigma.initialize()
    enigma.test()

    enigma.name =       '215 AAA FRA ABIRUXKP'   #VERIFIED
    enigma.url =        'http://www.mlb.co.jp/linux/science/genigma/enigma-referat/node4.html'
    enigma.plaintext =  'ANBULMEGRAZGOESTINGSTRENGGEHEIMEMELDUNG'
    enigma.ciphertext = 'PCDAONONEBCJBOGLYMEEYGSHRYUBUJHMJOQZLEX'
    enigma.rotorOrder =             ['II', 'I', 'V']
    enigma.ringList =               'AAA'
    enigma.startString =            'FRA'
    enigma.reflectorName =          'B'
    enigma.plugboardWiringString =  'AB IR UX KP'
    enigma.initialize()
    enigma.test()

    enigma.name =       'PAGE_40_AFDFX (Norrköping Enigma message)'   #VERIFIED
    enigma.url =        'https://enigma.hoerenberg.com/index.php?cat=Norrk%C3%B6ping%20messages&page=PAGE_40_AFDFX'
    enigma.plaintext =  'WAETIGKEITSBERIQTVOMXSEQSXDREIXVIERNULLXSEITEXZWOXLFLXKDOXEINSXXPREIXSTRIQXNULLXLFLXKDOXEINS'
    enigma.ciphertext = 'DSBDBDCOSGCYAMDSITWWZUSDLJBXAPVJMXLDKPUUIAPYYUSNRJOHKGOWRCDKBNYZELVKWYMWOPNTVOSNFNOCFNUDSQPU'
    enigma.rotorOrder =             ['V', 'I', 'IV']
    enigma.ringList =               'KSH'
    enigma.startString =            'SIN'
    enigma.reflectorName =          'B'
    enigma.plugboardWiringString =  'BF DV EJ GM IY KP LZ NX OQ TW'
    enigma.initialize()
    enigma.test()

    enigma.name =       'PAGE_40_AYDJM (Norrköping Enigma message)'   #VERIFIED
    enigma.url =        'https://enigma.hoerenberg.com/index.php?cat=Norrk%C3%B6ping%20messages&page=PAGE_40_AYDJM'
    enigma.plaintext =  'TAETIGKEITSBERIQTVOMXSEQSXDREIXVIERNULLXSONDERANLAGEXZWOXSXITEXZWOXLGXKDOXROEMDREIXXNULLXSTRIQXNULLXSEITEXDREIXLGXKDOXROEMXDREIXNULLXSTRIQXZWOXLGXKDOXROEMXDREI'
    enigma.ciphertext = 'YMZAXOZBCWGZFIGIMWXQWGAKOHLICKJOHFJFVNAVAUHAWSXEILABDJUBEUAQSAMWJWGUPOJZNLBCDDAVBXBQOPTYRNZTPUPABQBOJDJOXRJKMGLYJLQPBAQXAIAVXOGQHKTTOPHZROKUMSSJLONINRGUBQFAJPP'
    enigma.rotorOrder =             ['V', 'I', 'IV']
    enigma.ringList =               'KSH'
    enigma.startString =            'OEL'
    enigma.reflectorName =          'B'
    enigma.plugboardWiringString =  'BF DV EJ GM IY KP LZ NX OQ TW'
    enigma.initialize()
    enigma.test()

    enigma.name =       'PAGE_40_PYCMW (Norrköping Enigma message)' #NOT VERIFIED
    enigma.url =        'https://enigma.hoerenberg.com/index.php?cat=Norrk%C3%B6ping%20messages&page=PAGE_40_AYDJM'
    enigma.plaintext =  'TAETIGSKEITSBERIQTFUERDENSEQSXDREIXVIERNULLXSONDERANLAGEZWOXSEITENZAHLZWOXLGTKDOXROEMZWOXNEUNMOMMAFUENFXLGXKDOXROEMZWOX'
    enigma.ciphertext = 'PYZXKLLQFHELFTJAWURWBYZOWLFHFMNFYMWEGDECHKKGXHLTRQOEKIBFKLHHQWXLIQVMDVYTSZHXUSJXDACDTBBWUJIFEGIYZZJXIYHLVWNRVEVQKMGWNMH'
    enigma.rotorOrder =             ['I', 'III', 'V']
    enigma.ringList =               'XLG'
    enigma.startString =            'GRB'
    enigma.reflectorName =          'B'
    enigma.plugboardWiringString =  'AK BZ CE DX GL JQ MU NV OT SY'
    enigma.initialize()
    enigma.test()

    enigma.name =       'PAGE_40_ZBAQW (Norrköping Enigma message)'   #VERIFIED
    enigma.url =        'https://enigma.hoerenberg.com/index.php?cat=Norrk%C3%B6ping%20messages&page=PAGE_40_ZBAQW'
    enigma.plaintext =  'TAETIGKEITSBIRICHTVOMXSEQSXOREIXVIERNULLXAQTXAUFGENOMMENXNULLXMEFOERDERTXLEITSTELLEWARXKRAKAUXLXGXKDOXROEMXAQTX'
    enigma.ciphertext = 'OWTSBSTOLYMVABUFAAUWECRQYISIGZDZMCHECVVZGIZOYUJIDTFZUUOURDWVIMYOLXIOGNMKCYQQEYYTWITHHBYXAWAYXRTBUOOXXPNETERPZBI'
    enigma.rotorOrder =             ['I', 'III', 'V']
    enigma.ringList =               'XLG'
    enigma.startString =            'HJO'
    enigma.reflectorName =          'B'
    enigma.plugboardWiringString =  'AK BZ CE DX GL JQ MU NV OT SY'
    enigma.initialize()
    enigma.test()

    enigma.name =       'PAGE_47_RTQSX (Norrköping Enigma message)'   #VERIFIED
    enigma.url =        'https://enigma.hoerenberg.com/index.php?cat=Norrk%C3%B6ping%20messages&page=PAGE_47_RTQSX'
    enigma.plaintext =  'KOMMAHIERZUWIRDDERVERKGHRDERGEGENFUNKSTELVEBEOBAQTETUNDBXIABGABEEIFESFUERDIEABSTMMENDEFUMKSTELLENIQTWIQTVGENFUKKSPRUQESUSWPUNKTDEREMPFAENGERAUFSQWNBUNGSLUECKEX'
    enigma.ciphertext = 'WNLWSYZUKNCQSHOQCLAZLRJXDRIFLDWNMOQIBQZGJGIHRPTIWPHYAUCEIJMMYXPTONGQOXDAUKYIVNQAOYVFCXKJNOJQVPUMUZHFRMGGJIUILKUIESJKOOJNIYNBDVBOXMUAKREMODLBXHYGTXLUNFNHWYWQGVO'
    enigma.rotorOrder =             ['IV', 'I', 'V']
    enigma.ringList =               'PVG'
    enigma.startString =            'KJO'
    enigma.reflectorName =          'B'
    enigma.plugboardWiringString =  'AG BS CW EN FZ HJ IO KT LX PR'
    enigma.initialize()
    enigma.test()

    enigma.name =       'PAGE_47_KRNV  (Swedish Enigma M3 intercept)'   #VERIFIED
    enigma.url =        'https://enigma.hoerenberg.com/index.php?cat=Norrk%C3%B6ping%20messages&page=PAGE_47_KRNV'
    enigma.plaintext =  'SEENACHRICHTLEUCHTSCHIFFEMILSXBQPLANNORDFUENFNUOMERDREIEINSLIEGTWIEDERAUFSTATION'
    enigma.ciphertext = 'IWNOXNFTRJNPMRAWPZNKQECKKTWIOBEXLBHFLHHJLMYCZGCYSLCUDIJDLGQQPRYURGKGTFKRQEPIOVBI'
    enigma.rotorOrder =             ['VII', 'IV', 'VI']
    enigma.ringList =               'AGW'
    enigma.startString =            'RTA'
    enigma.reflectorName =          'B'
    enigma.plugboardWiringString =  'BM DX EW GP JO KV NZ RT'
    enigma.initialize()
    enigma.test()
    
    enigma.name =       'M4 Schroeder Message'    #VERIFIED
    enigma.url =        'https://enigma.hoerenberg.com/index.php?cat=M4%20Project%202006&page=Schroeder%20Message'
    enigma.plaintext =  'VVVJSCHREEDERJAUFGELEITKURSFUENFFUENFGRADNICHTSGEFUNDENYMARSCAIEREBEFOHLENESQUADRATXSTANRORTMARQUANTONJOTADREINEUNNEUNFUENFXSSSOOOVIERYSEEDREMYEINSNULYYEINSNULBEDECKTYZWOACHTMBSTEIGTYNBBELSICHTEINSSMT'
    enigma.ciphertext = 'TMKFNWZXFFIIYXUTIHWMDHXIFZEQVKDVMQSWBQNDYOZFTIWMJHXHYRPACZUGRREMVPANWXGTKTHNRLVHKZPGMNMVSECVCKHOINPLHHPVPXKMBHOKCCPDPEVXVVHOZZQBIYIEOUSEZNHJKWHYDAGTXDJDJKJPKCSDSUZTQCXJDVLPAMGQKKSHPHVKSVPCBUWZFIZPFUUP'
    enigma.rotorOrder =             ['BETA', 'II', 'IV', 'I']
    enigma.ringList =               'AANV'
    enigma.startString =            'MCSF'
    enigma.reflectorName =          'B THIN'
    enigma.plugboardWiringString =  'AT CL DH EP FG IO JN KQ MU RX'
    enigma.initialize()
    enigma.test()
    
    enigma.name =       'M4 Looks Message'    #VERIFIED
    enigma.url =        'https://enigma.hoerenberg.com/index.php?cat=M4%20Project%202006&page=Looks%20Message'
    enigma.plaintext =  'VONVONJLOOKSJHFFTTTEINSEINSDREIZWOYYQNNSNEUNINHALTXXBEIANGRIFFUNTERWASSERGEDRUECKTYWABOSXLETZTERGEGNERSTANDNULACHTDREINULUHRMARQUANTONJOTANEUNACHTSEYHSDREIYZWOZWONULGRADYACHTSMYSTOSSENACHXEKNSVIERMBFAELLTYNNNNNNOOOVIERYSICHTEINSNULL'
    enigma.ciphertext = 'NCZWVUSXPNYMINHZXMQXSFWXWLKJAHSHNMCOCCAKUQPMKCSMHKSEINJUSBLKIOSXCKUBHMLLXCSJUSRRDVKOHULXWCCBGVLIYXEOAHXRHKKFVDREWEZLXOBAFGYUJQUKGRTVUKAMEURBVEKSUHHVOYHABCJWMAKLFKLMYFVNRIZRVVRTKOFDANJMOLBGFFLEOPRGTFLVRHOWOPBEKVWMUQFMPWPARMFHAGKXIIBG'
    enigma.rotorOrder =             ['BETA', 'II', 'IV', 'I']
    enigma.ringList =               'AAAV'
    enigma.startString =            'VJNA'
    enigma.reflectorName =          'B THIN'
    enigma.plugboardWiringString =  'AT BL DF GJ HM NW OP QY RZ VX'
    enigma.initialize()
    enigma.test()

    enigma.name =       'M4 SOLUTION OF THE LAST OF THE “HMS HURRICANE” INTERCEPTS'    #VERIFIED
    enigma.url =        'https://enigma.hoerenberg.com/index.php?cat=M4%20Project%202006&page=Rasch%20Message'
    enigma.plaintext =  'BOOTKLARXBEIJSCHNOORBETWAZWOSIBENXNOVXSECHSNULCBMXPROVIANTBISZWONULXDEZXBENOETIGEGLMESERYNOCHVIEFKLHRXSTEHEMARQUBRUNOBRUNFZWOFUHFXLAGWWIEJKCHAEFERJXNNTWWWFUNFYEINSFUNFMBSTEIGENDYGUTESIWXDVVVJRASCH'
    enigma.ciphertext = 'HCEYZTCSOPUPPZDICQRDLWXXFACTTJMBRDVCJJMMZRPYIKHZAWGLYXWTMJPQUEFSZBOTVRLALZXWVXTSLFFFAUDQFBWRRYAPSBOWJMKLDUYUPFUQDOWVHAHCDWAUARSWTKOFVOYFPUFHVZFDGGPOOVGRMBPXXZCANKMONFHXPCKHJZBUMXJWXKAUODXZUCVCXPFT'
    enigma.rotorOrder =             ['BETA', 'VI', 'I', 'III']
    enigma.ringList =               'ZZTG'
    enigma.startString =            'NBHL'
    enigma.reflectorName =          'B THIN'
    enigma.plugboardWiringString =  'BQ CR DI EJ KW MT OS PX UZ GH'
    enigma.initialize()
    enigma.test()

    



    
